part of masamune.data;

/// Create a searchable collection.
/// 
/// By specifying [onSearch], you can query the API during the search and wait for the result.
/// During that time, the data will be initialized.
/// 
/// Start the search with the [search] method.
/// 
/// Others are the same as [MultiDataCollection].
class SearchableDataCollection extends MultiDataCollection {
  Future Function( String text, Iterable<IDataCollection> collections ) _onSearch;
  Iterable<String> _retrieveKeys;
  /// Process to create a new instance.
  /// 
  /// Do not use from outside the class.
  /// 
  /// [path]: Destination path.
  /// [isTemporary]: True if the data is temporary.
  @override
  @protected
  T createCloneInstance<T extends IClonable>( String path, bool isTemporary ) => SearchableDataCollection._(
    path:path,
    onSearch: this._onSearch,
    collections: this._collection,
    orderBy: this.orderBy,
    orderByKey: this.orderByKey,
    thenBy: this.thenBy,
    thenByKey: this.thenByKey
  ) as T;

  
  /// Create a searchable collection.
  /// 
  /// By specifying [onSearch], you can query the API during the search and wait for the result.
  /// During that time, the data will be initialized.
  /// 
  /// Start the search with the [search] method.
  /// 
  /// Others are the same as [MultiDataCollection].
  /// 
  /// [path]: Collection path.
  /// [collections]: The collection to store.
  /// [onSearch]: Search callback.
  /// [retrieveKeys]: Target key to search.
  /// [orderBy]: Sort order.
  /// [orderByKey]: Key for sorting.
  /// [thenBy]: Sort order when the first sort has the same value.
  /// [thenByKey]: Sort key when the first sort has the same value.
  factory SearchableDataCollection(
    String path, 
    {
      Iterable<IDataCollection> collections = const [], 
      Future onSearch( String text, Iterable<IDataCollection> collections ),
      Iterable<String> retrieveKeys = const [ "name", "text", "tag", "search" ],
      OrderBy orderBy = OrderBy.none, 
      OrderBy thenBy = OrderBy.none, 
      String orderByKey, String thenByKey
    }
  ){
    path = path?.trimString( Const.slash )
      ?.replaceAll( "//", "/" )
      ?.replaceAll(":/", "://")
      ?.applyTags();
    if( isEmpty(path) ){
      Log.error("The path is invalid.");
      return null;
    }
    SearchableDataCollection collection = PathMap.get<SearchableDataCollection>( path );
    if( collection != null ) {
      if( onSearch != null ) collection._onSearch = onSearch;
      if( retrieveKeys != null ) collection._retrieveKeys = retrieveKeys;
      if( orderBy != OrderBy.none ) collection.orderBy = orderBy;
      if( thenBy != OrderBy.none ) collection.thenBy = thenBy;
      if( isNotEmpty( orderByKey ) ) collection.orderByKey= orderByKey;
      if( isNotEmpty( thenByKey ) ) collection.thenByKey = thenByKey;
      if( collections != null ) {
        collection._collection = collections;
        collection._build();
      }
      return collection;
    }
    return SearchableDataCollection._(
      path:path, 
      onSearch: onSearch,
      retrieveKeys: retrieveKeys,
      collections:collections, 
      orderBy: orderBy, 
      orderByKey: orderByKey, 
      thenBy: thenBy, 
      thenByKey: thenByKey
    );
  }
  /// Create a searchable collection.
  /// 
  /// By specifying [onSearch], you can query the API during the search and wait for the result.
  /// During that time, the data will be initialized.
  /// 
  /// Start the search with the [search] method.
  /// 
  /// Others are the same as [MultiDataCollection].
  /// 
  /// [path]: Collection path.
  /// [collectionsFuture]: The collection to store.
  /// [collections]: The collection to store.
  /// [onSearch]: Search callback.
  /// [retrieveKeys]: Target key to search.
  /// [orderBy]: Sort order.
  /// [orderByKey]: Key for sorting.
  /// [thenBy]: Sort order when the first sort has the same value.
  /// [thenByKey]: Sort key when the first sort has the same value.
  static Future<SearchableDataCollection> fromFuture(
    {
      String path,
      Iterable<Future<IDataCollection>> collections = const [],
      Future onSearch( String text, Iterable<IDataCollection> collections ),
      Iterable<String> retrieveKeys = const [ "name", "text", "tag", "search" ],
      OrderBy orderBy = OrderBy.none,
      OrderBy thenBy = OrderBy.none,
      String orderByKey,
      String thenByKey
    }
  ) async {
    path = path?.trimString( Const.slash )
      ?.replaceAll( "//", "/" )
      ?.replaceAll(":/", "://")
      ?.applyTags();
    if( isNotEmpty(path) ) {
      SearchableDataCollection collection = PathMap.get<SearchableDataCollection>( path );
      if( collection != null ) {
        if( onSearch != null ) collection._onSearch = onSearch;
        if( retrieveKeys != null ) collection._retrieveKeys = retrieveKeys;
        if( orderBy != OrderBy.none ) collection.orderBy = orderBy;
        if( thenBy != OrderBy.none ) collection.thenBy = thenBy;
        if( isNotEmpty( orderByKey ) ) collection.orderByKey= orderByKey;
        if( isNotEmpty( thenByKey ) ) collection.thenByKey = thenByKey;
        if( collections != null ) {
          collection._collection = await collections?.mapAsync()?.toList();
          collection._build();
        }
        return collection;
      }
    }    
    List<IDataCollection> tmp = await collections?.mapAsync()?.toList();
    return SearchableDataCollection._(
      path:path,
      onSearch: onSearch,
      retrieveKeys: retrieveKeys,
      collections:tmp,
      orderBy: orderBy,
      orderByKey: orderByKey,
      thenBy: thenBy,
      thenByKey: thenByKey 
    );
  }
  SearchableDataCollection._(
    {
      String path,
      Iterable<IDataCollection> collections = const [],
      Future onSearch( String text, Iterable<IDataCollection> collections ),
      Iterable<String> retrieveKeys = const [ "name", "text", "tag", "search" ],
      OrderBy orderBy = OrderBy.none,
      OrderBy thenBy = OrderBy.none,
      String orderByKey,
      String thenByKey,
      bool isTemporary
    }
  ) :
  this._onSearch = onSearch,
  this._retrieveKeys = retrieveKeys,
  super._(
    path:path,
    collections: collections,
    orderBy: orderBy,
    thenBy: thenBy,
    orderByKey: orderByKey,
    thenByKey: thenByKey,
    isTemporary: isTemporary
  );
  @override
  void _build(){
    if( this._collection == null || this._collection.length <= 0 ) return;
    this.clear();
    this._filtering();
  }

  /// True while searching
  bool get isSearching => this._isSearching;
  bool _isSearching;

  /// Current search text
  String get searchText => this._searchText;

  /// Search the data.
  /// 
  /// [text]: Search text.
  Future<SearchableDataCollection> search( String text ) async {
    if( this._collection == null || this._collection.length <= 0 ) return this;
    try {
      this._isSearching = true;
      this._searchText = text;
      if( isEmpty( text ) ) {
        this._filtering();
        this._isSearching = false;
        return this;
      }
      this.clear();
      if( this._onSearch != null ){
        await this._onSearch( text, this._collection );
      }
      this._filtering();
    } catch( e ){
      Log.error(e.toString());
    }
    this._isSearching = false;
    return this;
  }
  void _filtering(){
    if( isEmpty( this._searchText ) ){
      for( IDataCollection collection in this._collection ){
        if( collection == null ) continue;
        for( IDataDocument document in collection ){
          if( document == null ) continue;
          this.setInternal(document);
        }
      }
    } else {
      List<String> search = this._searchText.toLowerCase().split(Const.space);
      for( IDataCollection collection in this._collection ){
        if( collection == null ) continue;
        for( IDataDocument document in collection ){
          if( document == null ) continue;
          Log.msg(search.length);
          if( search.length > 0 ){
            if( 
              !this._retrieveKeys.any(
                ( retrieve ) => search.any(
                  ( s ) => document.getString(retrieve)?.toLowerCase()?.contains( s ) ?? false
                )
              )
            ) continue;
          } 
          this.setInternal(document);
        }
      }
    }
    this.sort();
  }
  String _searchText;
}